"""Calculators module for AsusRouter"""

from __future__ import annotations

from asusrouter.const import(
    CONST_PERCENTS,
    CONST_ZERO,
    DATA_TOTAL,
    DATA_USAGE,
    DATA_USED,
    DEFAULT_USAGE_DIGITS,
    ERROR_ZERO_DIVISION,
)
from asusrouter.error import AsusRouterValueError

DEFAULT_USAGE_NONE = {
    DATA_TOTAL: 0,
    DATA_USED: 0,
}


def usage(current_used : (int | float), current_total : (int | float), previous_used : (int | float) = CONST_ZERO, previous_total : (int | float) = CONST_ZERO) -> float:
    """Calculate usage in percents"""

    if current_used == previous_used:
        return CONST_ZERO

    if current_total == previous_total:
        raise AsusRouterValueError(ERROR_ZERO_DIVISION.format("current_total == previous_total"))

    used = current_used - previous_used
    total = current_total - previous_total

    if used < 0:
        raise AsusRouterValueError("Usage cannot be negative, used = {}".format(used))
    if total < 0:
        raise AsusRouterValueError("Usage cannot be negative, total = {}".format(total))

    if used > total:
        raise AsusRouterValueError("Usage cannot be above 100%, used = {}, total = {}".format(used, total))

    return round(
        CONST_PERCENTS * (current_used - previous_used) / (current_total - previous_total)
    , DEFAULT_USAGE_DIGITS)


def usage_in_dict(after : dict[str, (int | float)], before : dict[str, (int | float)] = DEFAULT_USAGE_NONE) -> dict[str, (int | float)]:
    """Calculate usage in percents in a dictionary"""

    after[DATA_USAGE] = usage(after[DATA_USED], after[DATA_TOTAL], before[DATA_USED], before[DATA_TOTAL])
    
    return after


def speed(after : (int | float), before : (int | float), time_delta : (int | float) | None = None, overflow : (int | float) | None = None) -> float:
    """Calculate speed"""

    if time_delta is None:
        return CONST_ZERO
    elif time_delta == CONST_ZERO:
        raise AsusRouterValueError(ERROR_ZERO_DIVISION.format("time_delta"))

    diff = after - before

    # If we care about overflow values
    if overflow is not None:
        if (diff < CONST_ZERO):
            diff += overflow

    return diff / time_delta





"""Compilers module for AsusRouter"""

from __future__ import annotations

from asusrouter.const import(
    AR_HOOK_TEMPLATE,
    KEY_NVRAM_GET,
)


def hook(commands : dict[str, str] | None = None) -> str:
    """"Hook compiler"""

    data = str()
    if commands is not None:
        for item in commands:
            data += AR_HOOK_TEMPLATE.format(item, commands[item])

    return data


def nvram(values : list[str] | str | None = None) -> str:
    """NVRAM request compiler"""

    if values is None:
        return str()

    if type(values) == str:
        return AR_HOOK_TEMPLATE.format(KEY_NVRAM_GET, values)

    request = str()
    for value in values:
        request += AR_HOOK_TEMPLATE.format(KEY_NVRAM_GET, value)

    return request





"""AsusRouter module"""

from __future__ import annotations

import logging
_LOGGER = logging.getLogger(__name__)

import asyncio
from datetime import datetime
from typing import Any

from asusrouter import (
    AsusDevice,
    Connection,
    Monitor,
    AsusRouterIdentityError,
    AsusRouterServerDisconnectedError,
    AsusRouterServiceError,
)
from asusrouter.util import calculators, compilers, converters, parsers
from asusrouter.const import (
    AR_DEVICE_IDENTITY,
    AR_HOOK_DEVICES,
    AR_KEY_CPU,
    AR_KEY_DEVICES,
    AR_KEY_LED,
    AR_KEY_NETWORK,
    AR_KEY_RAM,
    AR_KEY_SERVICE_COMMAND,
    AR_KEY_SERVICE_MODIFY,
    AR_KEY_SERVICE_REPLY,
    AR_KEY_WAN,
    AR_PATH,
    AR_SERVICE_COMMAND,
    AR_SERVICE_CONTROL,
    AR_SERVICE_DROP_CONNECTION,
    DATA_BY_CORE,
    DATA_TOTAL,
    DATA_USAGE,
    DATA_USED,
    DEFAULT_ACTION_MODE,
    DEFAULT_CACHE_TIME,
    DEFAULT_SLEEP_TIME,
    ERROR_IDENTITY,
    ERROR_SERVICE,
    ERROR_SERVICE_UNKNOWN,
    INTERFACE_TYPE,
    KEY_ACTION_MODE,
    KEY_CPU,
    KEY_HOOK,
    KEY_NETWORK,
    KEY_RAM,
    KEY_WAN,
    MONITOR_MAIN,
    MSG_ERROR,
    MSG_INFO,
    MSG_SUCCESS,
    NVRAM_LIST,
    PORT_TYPE,
)


class AsusRouter:
    """The interface class"""

    def __init__(
        self,
        host : str,
        username : str | None = None,
        password : str | None = None,
        port : int | None = None,
        use_ssl : bool = False,
        cert_check : bool = True,
        cert_path : str = "",
        cache_time : int = DEFAULT_CACHE_TIME,
        enable_monitor : bool = True,
        enable_control : bool = False):
        """Init"""

        self._host = host

        self._enable_monitor : bool = enable_monitor
        self._enable_control : bool = enable_control

        self._cache_time : int = cache_time


        self._device_cpu_cores : list[int] | None = None
        self._device_ports : dict[str, str] | None = None
        self._device_boottime : datetime | None = None

        self._monitor_main : Monitor = Monitor()
        self._monitor_nvram : Monitor = Monitor()
        self._monitor_misc : Monitor = Monitor()
        self._monitor_devices : Monitor = Monitor()

        self._identity : AsusDevice | None = None
        self._status_led : bool = False


        """Connect"""
        self.connection = Connection(host = self._host, username = username, password = password, port = port, use_ssl = use_ssl, cert_check = cert_check, cert_path = cert_path)


    async def async_identify(self) -> None:
        """Identify the device"""

        _LOGGER.debug(MSG_INFO["identifying"])

        # Compile
        query = list()
        for item in AR_DEVICE_IDENTITY:
            query.append(item)


        # Collect data
        message = compilers.nvram(query)
        try:
            raw = await self.async_hook(message, force = True)
        except Exception as ex:
            AsusRouterIdentityError(ERROR_IDENTITY.format(self._host, str(ex)))

        # Parse
        identity = dict()
        for item in AR_DEVICE_IDENTITY:
            key = item.get()
            try:
                data = item.method(raw[item.value]) if item.method else raw[item.value]
                if key in identity:
                    if type(identity[key]) is list:
                        identity[key].extend(data)
                    else:
                        identity[key] = data
                else:
                    identity[key] = data
            except Exception as ex:
                AsusRouterIdentityError(ERROR_IDENTITY.format(self._host, str(ex)))

        # Save static values
        self._status_led = raw[AR_KEY_LED]

        self._identity = AsusDevice(**identity)

        _LOGGER.debug(MSG_SUCCESS["identity"])


    async def async_get_identity(self, force : bool = False) -> AsusDevice:
        """Return device identity"""

        if (not self._identity
            or force
        ):
            await self.async_identify()

        return self._identity


    async def async_connect(self) -> bool:
        """Connect to the device"""

        try:
            await self.connection.async_connect()
        except Exception as ex:
            raise ex

        await self.async_identify()

        return True


    async def async_drop_connection(self) -> bool:
        """Drop connection when device will not reply because of our actions"""

        try:
            await self.connection.async_drop_connection()
        except Exception as ex:
            raise ex

        return True


    async def async_disconnect(self) -> bool:
        """Disconnect from the device"""

        try:
            await self.connection.async_disconnect()
        except Exception as ex:
            raise ex

        return True


    async def async_hook(self, hook : str | None = None, force : bool = False) -> dict[str, Any]:
        """Hook data from device"""

        result = {}

        if (not self._enable_monitor
            and not force
        ):
            _LOGGER.error(MSG_ERROR["disabled_monitor"])
            return result

        if hook is None:
            _LOGGER.debug(MSG_INFO["empty_request"])
            return result

        try:
            result = await self.connection.async_run_command(command = "{}={}".format(KEY_HOOK, hook))
            _LOGGER.debug("{}: {}".format(MSG_SUCCESS["hook"], hook))
        except Exception as ex:
            raise ex

        # Check for errors during hook
        if self.connection.error:
            _LOGGER.debug(MSG_INFO["error_flag"])
            await self.async_handle_error()

        return result


    async def async_command(self, commands : dict[str, str], action_mode : str = DEFAULT_ACTION_MODE) -> dict[str, Any]:
        """Command device to run a service or set parameter"""

        result = {}

        if not self._enable_control:
            _LOGGER.error(MSG_ERROR["disabled_control"])
            return result

        request : dict = {
            KEY_ACTION_MODE: action_mode,
        }
        for command in commands:
            request[command] = commands[command]

        try:
            result = await self.connection.async_run_command(command = str(request), endpoint = AR_PATH["command"])
            _LOGGER.debug("{}: {}".format(MSG_SUCCESS["command"], request))
        except Exception as ex:
            raise ex

        return result


    async def async_load(self, page : str | None = None) -> dict[str, Any]:
        """Return the data from the page"""

        result = {}

        if not self._enable_monitor:
            _LOGGER.error(MSG_ERROR["disabled_monitor"])
            return result

        if page is None:
            _LOGGER.debug(MSG_INFO["empty_request"])
            return result

        try:
            result = await self.connection.async_run_command(command = "", endpoint = page)
            _LOGGER.debug(MSG_SUCCESS["load"].format(page))
        except Exception as ex:
            _LOGGER.error(ex)

        return result


    ### MONITORS -->


    async def async_monitor_main(self) -> None:
        """Get all the main monitor values. Non-cacheable"""

        while self._monitor_main.active:
            _LOGGER.debug(MSG_INFO["monitor_sleep"].format("main"))
            await asyncio.sleep(DEFAULT_SLEEP_TIME)
            _LOGGER.debug(MSG_INFO["monitor_wakeup"].format("main"))
            return

        self._monitor_main.start()

        monitor_main = Monitor()

        hook = compilers.hook(MONITOR_MAIN)
        data = await self.async_hook(hook)

        monitor_main.reset()

        ### CPU ###
        ## Not yet known what exactly is type of data. But this is the correct way to calculate usage from it

        # Keep safe
        if AR_KEY_CPU in data:

            raw_cpu = data[AR_KEY_CPU]

            # Check which cores CPU has or find it out and save for later
            if self._device_cpu_cores is None:
                self._device_cpu_cores = parsers.cpu_cores(raw = raw_cpu)

            # Traffic data from the device
            monitor_main[KEY_CPU] = parsers.cpu_usage(raw = raw_cpu, cores = self._device_cpu_cores)

            # Calculate actual usage in percents and save it. Only if there was old data for CPU
            if (self._monitor_main.ready
                and KEY_CPU in self._monitor_main
            ):
                for item in monitor_main[KEY_CPU]:
                    if item in self._monitor_main[KEY_CPU]:
                        monitor_main[KEY_CPU][item] = calculators.usage_in_dict(after = monitor_main[KEY_CPU][item], before = self._monitor_main[KEY_CPU][item])

        # Keep last data
        elif (self._monitor_main.ready
            and KEY_CPU in self._monitor_main
        ):
            monitor_main[KEY_CPU] = self._monitor_main[KEY_CPU]


        ### RAM ###
        ## Data is in KiB. To get MB as they are shown in the device Web-GUI, should be devided by 1024 (yes, those will be MiB)

        if AR_KEY_RAM in data:
            # Populate RAM with known values
            monitor_main[KEY_RAM] = parsers.ram_usage(raw = data[AR_KEY_RAM])

            # Calculate usage in percents
            if (DATA_USED in monitor_main[KEY_RAM]
                and DATA_TOTAL in monitor_main[KEY_RAM]
            ):
                monitor_main[KEY_RAM] = calculators.usage_in_dict(after = monitor_main[KEY_RAM])
        # Keep last data
        elif (self._monitor_main.ready
            and KEY_RAM in self._monitor_main
        ):
            monitor_main[KEY_RAM] = self._monitor_main[KEY_RAM]


        ### NETWORK ###
        ## Data in Bytes for traffic and bits/s for speeds

        if AR_KEY_NETWORK in data:
            # Calculate RX and TX from the HEX values. If there is no current value, but there was one before, get it from storage. Traffic resets only on device reboot or when above the limit. Device disconnect / reconnect does NOT reset it
            monitor_main[KEY_NETWORK] = parsers.network_usage(raw = data[AR_KEY_NETWORK])

            if (self._monitor_main.ready
                and KEY_NETWORK in self._monitor_main
            ):
                # Calculate speeds
                time_delta = (monitor_main.time - self._monitor_main.time).total_seconds()
                monitor_main[KEY_NETWORK] = parsers.network_speed(after = monitor_main[KEY_NETWORK], before = self._monitor_main[KEY_NETWORK], time_delta = time_delta)

        # Keep last data
        elif (self._monitor_main.ready
            and KEY_NETWORK in self._monitor_main
        ):
            monitor_main[KEY_NETWORK] = self._monitor_main[KEY_NETWORK]


        ### WAN STATE ###

        if AR_KEY_WAN in data:
            # Populate WAN with known values
            monitor_main[KEY_WAN] = parsers.wan_state(raw = data[AR_KEY_WAN])


        # Keep last data
        elif (self._monitor_main.ready
            and KEY_WAN in self._monitor_main
        ):
            monitor_main[KEY_WAN] = self._monitor_main[KEY_WAN]

        monitor_main.finish()
        self._monitor_main = monitor_main

        return


    async def async_monitor_nvram(self, groups : list[str] | str | None = None) -> None:
        """Get the NVRAM values for the specified group list"""

        while self._monitor_nvram.active:
            _LOGGER.debug(MSG_INFO["monitor_sleep"].format("nvram"))
            await asyncio.sleep(DEFAULT_SLEEP_TIME)
            _LOGGER.debug(MSG_INFO["monitor_wakeup"].format("nvram"))
            return

        self._monitor_nvram.start()

        monitor_nvram = Monitor()

        # If none groups were sent, will return all the known NVRAM values
        if groups is None:
            groups = [*NVRAM_LIST]

        # If only one group is sent
        if type(groups) is not list:
            groups = [groups.upper()]

        requests = []
        for group in groups:
            group = group.upper()
            if group in NVRAM_LIST:
                for value in NVRAM_LIST[group]:
                    requests.append(value)
            else:
                _LOGGER.warning("There is no {} in known NVRAM groups".format(group))

        message = compilers.nvram(requests)
        result = await self.async_hook(message)

        monitor_nvram.reset()

        for item in result:
            monitor_nvram[item] = result[item]
        
        monitor_nvram.finish()
        self._monitor_nvram = monitor_nvram

        return


    async def async_monitor_misc(self) -> None:
        """Get all other useful values"""

        while self._monitor_misc.active:
            _LOGGER.debug(MSG_INFO["monitor_sleep"].format("misc"))
            await asyncio.sleep(DEFAULT_SLEEP_TIME)
            _LOGGER.debug(MSG_INFO["monitor_wakeup"].format("misc"))
            return

        self._monitor_misc.start()

        monitor_misc = Monitor()

        ### PORTS

        # Receive ports number and status (disconnected, 100 Mb/s, 1 Gb/s)
        data = await self.async_load(page = AR_PATH["ports"])

        monitor_misc.reset()

        monitor_misc["PORTS"] = dict()
        if "portSpeed" in data:
            for type in PORT_TYPE:
                monitor_misc["PORTS"][type] = dict()
            for value in data["portSpeed"]:
                for type in PORT_TYPE:
                    if type in value:
                        number = value.replace(type, "")
                        monitor_misc["PORTS"][type][number] = parsers.port_speed(data["portSpeed"][value])
                        break

        # Load devicemap
        data = await self.async_load(page = AR_PATH["devicemap"])
        monitor_misc["DEVICEMAP"] = data

        # Calculate boot time. Since precision is 1 second, could be that old and new are 1 sec different. In this case, we should not change the boot time, but keep the previous value to avoid regular changes
        if "SYS" in monitor_misc["DEVICEMAP"]:
            if "uptimeStr" in monitor_misc["DEVICEMAP"]["SYS"]:
                time = parsers.uptime(monitor_misc["DEVICEMAP"]["SYS"]["uptimeStr"])
                timestamp = int(time.timestamp())
                if not "BOOTTIME" in monitor_misc:
                    monitor_misc["BOOTTIME"] = dict()

                if "BOOTTIME" in self._monitor_misc:
                    _timestamp = self._monitor_misc["BOOTTIME"]["timestamp"]
                    _time = datetime.fromtimestamp(_timestamp)
                    if (time == _time
                        or abs(timestamp - _timestamp) < 2
                    ):
                        monitor_misc["BOOTTIME"] = self._monitor_misc["BOOTTIME"]
                    # This happens on reboots if data is checked before the device got correct time from NTP
                    elif timestamp - _timestamp < 0:
                        # Leave the same boot time, since we don't know the new correct time
                        monitor_misc["BOOTTIME"] = self._monitor_misc["BOOTTIME"]
                    else:
                        monitor_misc["BOOTTIME"]["timestamp"] = timestamp
                        monitor_misc["BOOTTIME"]["ISO"] = time.isoformat()
                else:
                    monitor_misc["BOOTTIME"]["timestamp"] = timestamp
                    monitor_misc["BOOTTIME"]["ISO"] = time.isoformat()

                if self._device_boottime != monitor_misc["BOOTTIME"]["ISO"]:
                    self._device_boottime = monitor_misc["BOOTTIME"]["ISO"]

        monitor_misc.finish()
        self._monitor_misc = monitor_misc

        return


    async def async_monitor_devices(self) -> None:
        """Monitor connected devices"""

        while self._monitor_devices.active:
            _LOGGER.debug(MSG_INFO["monitor_sleep"].format("devices"))
            await asyncio.sleep(DEFAULT_SLEEP_TIME)
            _LOGGER.debug(MSG_INFO["monitor_wakeup"].format("devices"))
            return

        self._monitor_devices.start()
        monitor_devices = Monitor()

        data = await self.async_hook(AR_HOOK_DEVICES)
        monitor_devices.reset()

        # Search for new data
        if AR_KEY_DEVICES in data:
            data = data[AR_KEY_DEVICES]
            for mac in data:
                if converters.is_mac_address(mac):
                    monitor_devices[mac] = parsers.connected_device(data[mac])
        # Or keep last data
        elif self._monitor_devices.ready:
            monitor_devices = self._monitor_devices

        monitor_devices.finish()
        self._monitor_devices = monitor_devices

        return


    ### <-- MONITORS


    async def async_handle_error(self) -> None:
        """Actions to be taken on connection errors"""

        # Most of errors come from device reboots. Handle it
        await self.async_handle_reboot()

        # Clear error
        await self.connection.async_reset_error()

        return


    async def async_handle_reboot(self) -> None:
        """Actions to be taken on reboot"""

        # Clear main monitor to prevent calculation errors in it
        if self._monitor_main.ready:
            self._monitor_main = Monitor()

        return


    async def async_find_interfaces(self, use_cache : bool = True) -> None:
        """Find available interfaces/type dictionary"""
        
        if (not self._monitor_nvram.ready
            or use_cache == False
        ):
            await self.async_monitor_nvram()

        ports = {}

        data = self._monitor_nvram
        
        for if_type in INTERFACE_TYPE:
            if if_type in data:
                values = data[if_type].split(" ")
                for item in values:
                    ports[item] = INTERFACE_TYPE[if_type]

        self._device_ports = ports

        return


    async def async_initialize(self):
        """Get all the data needed at the startup"""

        await self.async_monitor_main()
        await self.async_monitor_nvram()
        await self.async_find_interfaces()
        await self.async_monitor_misc()
        await self.async_monitor_devices()


    ### RETURN DATA -->


    async def async_get_devices(self, use_cache : bool = True) -> dict[str, Any]:
        """Return device list"""
        
        now = datetime.utcnow()
        if (not self._monitor_devices.ready
            or use_cache == False
            or (use_cache == True
                and self._cache_time < (now - self._monitor_devices.time).total_seconds()
            )
        ):
            await self.async_monitor_devices()

        return self._monitor_devices.copy()


    async def async_get_cpu(self, use_cache : bool = True) -> dict[str, float]:
        """Return CPU usage"""
        
        now = datetime.utcnow()
        if (not self._monitor_main.ready
            or not KEY_CPU in self._monitor_main
            or use_cache == False
            or (use_cache == True
                and self._cache_time < (now - self._monitor_main.time).total_seconds()
            )
            or self._device_cpu_cores is None
        ):
            await self.async_monitor_main()

        result = dict()

        for core in self._device_cpu_cores:
            if DATA_USAGE in self._monitor_main[KEY_CPU][core]:
                result[DATA_BY_CORE.format(core)] = self._monitor_main[KEY_CPU][core][DATA_USAGE]
        if DATA_USAGE in self._monitor_main[KEY_CPU][DATA_TOTAL]:
            result[DATA_TOTAL] = self._monitor_main[KEY_CPU][DATA_TOTAL][DATA_USAGE]

        return result


    async def async_get_cpu_labels(self) -> list[str]:
        """Return list of CPU cores"""

        if self._device_cpu_cores is None:
            await self.async_monitor_main()

        result = list()
        result.append(DATA_TOTAL)
        for core in self._device_cpu_cores:
            result.append(DATA_BY_CORE.format(core))

        return result


    async def async_get_ram(self, use_cache : bool = True) -> dict[str, (int | float)]:
        """Return RAM and its usage"""
        
        now = datetime.utcnow()
        if (not self._monitor_main.ready
            or use_cache == False
            or (use_cache == True
                and self._cache_time < (now - self._monitor_main.time).total_seconds()
            )
        ):
            await self.async_monitor_main()

        result = dict()

        # Check if RAM was monitored
        if (not self._monitor_main.ready
            or not KEY_RAM in self._monitor_main
        ):
            return result

        for value in self._monitor_main[KEY_RAM]:
            result[value] = self._monitor_main[KEY_RAM][value]

        return result


    async def async_get_ram_labels(self) -> list[str]:
        """Return list of CPU cores"""

        if not self._monitor_main.ready:
            await self.async_monitor_main()

        result = list()
        for value in self._monitor_main[KEY_RAM]:
            result.append(value)

        return result


    async def async_get_network(self, use_cache : bool = True) -> dict[str, (int | float)]:
        """Return traffic and speed for all interfaces"""
        
        now = datetime.utcnow()
        if (not self._monitor_main.ready
            or use_cache == False
            or (use_cache == True
                and self._cache_time < (now - self._monitor_main.time).total_seconds()
            )
        ):
            await self.async_monitor_main()

        result = dict()

        # Check if network was monitored
        if (not self._monitor_main.ready
            or not KEY_NETWORK in self._monitor_main
        ):
            return result

        for interface in self._monitor_main[KEY_NETWORK]:
            for value in self._monitor_main[KEY_NETWORK][interface]:
                result["{}_{}".format(interface, value)] = self._monitor_main[KEY_NETWORK][interface][value]

        return result


    async def async_get_network_labels(self) -> list[str]:
        """Return list of network interfaces"""

        if not self._monitor_main.ready:
            await self.async_monitor_main()
        if not self._identity:
            await self.async_identify()

        result = list()

        for interface in self._monitor_main[KEY_NETWORK]:
            result.append(interface)

        if (not "USB" in result
            and "dualwan" in self._identity.services
        ):
            result.append("USB")

        return result


    async def async_get_ports(self, use_cache : bool = True) -> dict[str, dict[str, int]]:
        """Return WAN/LAN ports status"""
        
        now = datetime.utcnow()
        if (not self._monitor_misc.ready
            or use_cache == False
            or (use_cache == True
                and self._cache_time < (now - self._monitor_misc.time).total_seconds()
            )
        ):
            await self.async_monitor_misc()

        return self._monitor_misc["PORTS"]


    async def async_get_wan(self, use_cache : bool = True) -> dict[str, str]:
        """Return WAN and its usage"""
        
        now = datetime.utcnow()
        if (not self._monitor_main.ready
            or use_cache == False
            or (use_cache == True
                and self._cache_time < (now - self._monitor_main.time).total_seconds()
            )
        ):
            await self.async_monitor_main()

        result = dict()

        # Check if WAN was monitored
        if (not self._monitor_main.ready
            or not KEY_WAN in self._monitor_main
        ):
            return result

        for value in self._monitor_main[KEY_WAN]:
            result[value] = self._monitor_main[KEY_WAN][value]

        return result


    ### <-- RETURN DATA


    ### SERVICES -->


    async def async_service_run(self, service : str, arguments : dict[str, Any] | None = None, drop_connection : bool = False) -> bool:
        """Generic service to run"""

        commands = {
            AR_KEY_SERVICE_COMMAND: service,
        }

        if arguments:
            commands.update(arguments)

        try:
            result = await self.async_command(commands = commands)
        except AsusRouterServerDisconnectedError as ex:
            # For services, that will block any further connections
            if drop_connection:
                _LOGGER.debug(MSG_INFO["drop_connection_service"].format(service))
                await self.async_drop_connection()
                return True
            else:
                raise ex

        if (not AR_KEY_SERVICE_REPLY in result
            or result[AR_KEY_SERVICE_REPLY] != service
            or not AR_KEY_SERVICE_MODIFY in result
        ):
            raise AsusRouterServiceError(ERROR_SERVICE.format(service, result))

        _LOGGER.debug(MSG_INFO["service"].format(service, arguments, result))

        return converters.bool_from_any(result[AR_KEY_SERVICE_MODIFY])


    async def async_service_led_set(self, value : bool | int | str) -> bool:
        """Set status of the LEDs"""

        value_to_set = converters.bool_from_any(value)

        service = "start_ctrl_led"
        arguments = {
            AR_KEY_LED: converters.int_from_bool(value_to_set),
        }
        result = await self.async_service_run(service = service, arguments = arguments)
        if result:
            self._status_led = value_to_set
        return result


    async def async_service_led_get(self) -> bool | None:
        """Return status of the LEDs"""

        key = AR_KEY_LED
        raw = await self.async_hook(compilers.nvram(key))
        if not converters.none_or_str(raw[key]):
            return None
        value = converters.bool_from_any(raw[key])
        self._status_led = value
        return value


    async def async_service_reboot(self) -> bool:
        """Reboot the device"""

        service = "reboot"
        await self.async_service_run(service = service, drop_connection = True)
        return True


    async def async_service_control(self, target : str, mode : str) -> bool:
        """Start / stop / (force) restart service"""

        if (not target in AR_SERVICE_CONTROL
            or not mode in AR_SERVICE_CONTROL[target]
        ):
            raise AsusRouterServiceError(ERROR_SERVICE_UNKNOWN.format(target, mode))

        service = AR_SERVICE_COMMAND[mode].format(target)

        drop_connection = False
        if target in AR_SERVICE_DROP_CONNECTION:
            drop_connection = True

        return await self.async_service_run(service = service, drop_connection = drop_connection)


    ### <-- SERVICES


    @property
    def connected(self) -> bool:
        return self.connection.connected


    @property
    def boottime(self) -> datetime:
        return self._device_boottime


    @property
    def led(self) -> bool:
        return self._status_led





"""Convertors module for AsusRouter"""

from __future__ import annotations

from datetime import datetime, timedelta, timezone
import re

from asusrouter.error import AsusRouterValueError

BOOL_FALSE : tuple[str, ...] = ("false", "block", "0", "off", )
BOOL_TRUE : tuple[str, ...] = ("true", "allow", "1", "on", )
ERROR_VALUE = "Wrong value: {} with original exception: {}"
ERROR_VALUE_TYPE = "Wrong value {} of type {}"
REGEX_MAC = re.compile("^([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})")


def int_from_str(raw : str, base : int = 10) -> int:
    """Convert string to integer"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    raw = raw.strip()

    if raw == str():
        return int(0)

    return int(raw, base = base)


def float_from_str(raw : str) -> float:
    """Convert striing to float"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    raw = raw.strip()

    if raw == str():
        return float(0)
    # For devices reporrting speeds with M at the end for mega
    if raw[-1] == "M":
        raw = raw[:-1]

    try:
        value = float(raw)
    except ValueError as ex:
        raise(AsusRouterValueError(ERROR_VALUE.format(raw, str(ex))))

    return value


def bool_from_any(raw : str | int | float) -> bool:
    """Converts string or number to bool"""

    _type = type(raw)

    if (_type == int
        or _type == float
    ):
        if raw == 0:
            return False
        else:
            return True

    elif _type == str:
        if raw.lower().strip() in BOOL_FALSE:
            return False
        elif raw.lower().strip() in BOOL_TRUE:
            return True
        else:
            raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    else:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))


def int_from_bool(raw : bool) -> int:
    """Convert boolean value to 0 or 1"""

    if type(raw) != bool:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    return 1 if raw else 0


def none_or_str(raw : str) -> str | None:
    """Returns either string or None if string is empty"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    if raw.strip() == str():
        return None

    return raw.strip()


def exists_or_not(raw : str) -> bool:
    """Returns whether value exists or is empty"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    if raw.strip() == str():
        return False

    return True


def timedelta_long(raw : str) -> timedelta:
    """Transform connection timedelta of the device to a proper datetime object when the device was connected"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
        
    raw = raw.strip()

    if raw == str():
        return timedelta()

    part = raw.split(":")
    try:
        return timedelta(hours = int(part[-3]), minutes = int(part[-2]), seconds = int(part[-1]))
    except ValueError as ex:
        raise(AsusRouterValueError(ERROR_VALUE.format(raw, str(ex))))


def time_from_delta(raw : str) -> datetime:
    """Transform time delta to the date in the past"""

    return datetime.utcnow().replace(microsecond = 0, tzinfo = timezone.utc) - timedelta_long(raw)


def is_mac_address(raw : str) -> bool:
    """Checks if string is MAC address"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    raw = raw.strip()

    if raw == str():
        return False

    if re.search(REGEX_MAC, raw):
        return True

    return False


def service_support(raw : str) -> list[str]:
    """Get the list of the supported services"""

    if type(raw) != str:
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    services = list()

    if raw == str():
        return services

    services = raw.split(" ")
    services = [i for i in services if i]

    return services





"""Dataclass module for AsusRouter"""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime

from asusrouter.util.converters import none_or_str


@dataclass
class ConnectedDevice:
    """Connected device class"""

    name : str | None = None
    mac : str | None = None

    ip : str | None = None
    ip_method : str | None = None

    internet_state : bool | None = None
    internet_mode : bool | None = None

    connection_type : str | None = None

    # WLAN only values
    online : bool | None = None
    rssi : int | None = None
    connected_since : int | None = None
    rx_speed : float | None = None
    tx_speed : float | None = None


@dataclass
class AsusDevice:
    """Asus device class"""

    serial : str | None = None
    mac : str | None = None
    model : str | None = None
    brand : str = "ASUSTek"
    fw_major : str | None = None
    fw_minor : str | None = None
    fw_build : str | None = None
    services : str | None = None
    led : bool = False


    def firmware(self) -> str:
        return "{}.{}_{}".format(self.fw_major, self.fw_minor, self.fw_build)


@dataclass
class Key:
    """Key class"""

    value : str
    value_to_use : str = ""
    method : function = none_or_str


    def __str__(self) -> str:
        """Return only `value` as default"""

        return self.value


    def get(self) -> str:
        """
        Get the proper value

        Returns
        -----
        `value_to_use` if exists, `value` otherwise
        """

        if self.value_to_use != "":
            return self.value_to_use

        return self.value


@dataclass
class Monitor(dict):
    """
    Monitor class

    In general this is dict with additions

    Properties
    -----
    `active`: bool flag of monitor being active

    `time`: datetime object showing the last time monitor was updated

    `ready`: bool flag if monitor was ever loaded

    Methods
    -----
    `start`: set `active` to True

    `stop`: set `active` to False

    `reset`: set `time` to utcnow()

    `finish`: set `ready` to True
    """

    active : bool = False
    time : datetime | None = None
    ready : bool = False


    def start(self) -> None:
        """Set to active"""

        self.active = True


    def stop(self) -> None:
        """Set to not-active"""

        self.active = False


    def reset(self) -> None:
        """Reset time to utcnow"""

        self.time = datetime.utcnow()


    def finish(self) -> None:
        """Set ready status to True"""

        self.ready = True





"""AsusRouter constants module"""

from asusrouter.dataclass import Key
from asusrouter.util.converters import bool_from_any, exists_or_not, int_from_str, float_from_str, time_from_delta, service_support

#Use the last known working Android app user-agent, so the device will reply
#AR_USER_AGENT = "asusrouter-Android-DUTUtil-1.0.0.255"
#Or use just "asusrouter--DUTUtil-", since only this is needed for a correct replies
AR_USER_AGENT = "asusrouter--DUTUtil-"
#Or even this - all the response will be correct, but the HTTP header will be missing 'AiHOMEAPILevel', 'Httpd_AiHome_Ver' and 'Model_Name' on connect
#AR_USER_AGENT = "asusrouter--"


AR_API = [
    "Model_Name",
    "AiHOMEAPILevel",
    "Httpd_AiHome_Ver",
]

CONST_BITSINBYTE = 8
CONST_PERCENTS = 100
CONST_ZERO = 0.0

DATA_FREE = "free"
DATA_TOTAL = "total"
DATA_USAGE = "usage"
DATA_USED = "used"
DATA_RX = "rx"
DATA_TX = "tx"
DATA_TRAFFIC = [DATA_RX, DATA_TX]

DATA_BY_CORE = "core_{}"
DATA_ADD_SPEED = "{}_speed"

DEFAULT_TRAFFIC_OVERFLOW = 4294967296

KEY_CPU = "CPU"
KEY_RAM = "RAM"
KEY_NETWORK = "NETWORK"
KEY_WAN = "WAN"

AR_DEVICE_IDENTITY : tuple[Key, ...] = (
    Key("serial_no", "serial"),
    Key("label_mac", "mac"),
    Key("model", "model"),
    Key("firmver", "fw_major"),
    Key("buildno", "fw_minor"),
    Key("extendno", "fw_build"),
    Key("rc_support", "services", method = service_support),
    Key("ss_support", "services", method = service_support),
    Key("led_val", "led", method = exists_or_not)
)

AR_DEVICE_ATTRIBUTES_LIST : tuple[Key, ...] = (
    Key("mac", "name"),
    Key("name", "name"),
    Key("nickName", "name"),
    Key("mac"),
    Key("ip"),
    Key("ipMethod", "ip_method"),
    Key("internetState", "internet_state", method = bool_from_any),
    Key("internetMode", "internet_mode", method = bool_from_any),
    Key("isWL", "connection_type", method = int_from_str),
    Key("isOnline", "online", method = bool_from_any),
    Key("rssi", method = int_from_str),
    Key("wlConnectTime", "connected_since", time_from_delta),
    Key("curRx", "rx_speed", method = float_from_str),
    Key("curTx", "tx_speed", method = float_from_str),
)
AR_KEY_CPU = "cpu_usage"
AR_KEY_CPU_ITEM = "cpu{}_{}"
AR_KEY_CPU_LIST : tuple[Key, ...] = (
    Key(DATA_TOTAL),
    Key(DATA_USAGE, DATA_USED)
)
AR_KEY_DEVICE_NAME_LIST = ["nickName", "name", "mac"]
AR_KEY_DEVICEMAP = "devicemap"
AR_KEY_DEVICES = "get_clientlist"
AR_KEY_DEVICES_LIST = "maclist"
AR_KEY_LED = "led_val"
AR_KEY_NETWORK = "netdev"
AR_KEY_NETWORK_ITEM = "{}_{}"
AR_KEY_RAM = "memory_usage"
AR_KEY_RAM_ITEM = "mem_{}"
AR_KEY_RAM_LIST = [DATA_FREE, DATA_TOTAL, DATA_USED]
AR_KEY_SERVICE_COMMAND = "rc_service"
AR_KEY_SERVICE_MODIFY = "modify"
AR_KEY_SERVICE_REPLY = "run_service"


AR_KEY_NETWORK_GROUPS = {
    "INTERNET" : "WAN",     # main WAN
    "INTERNET1" : "USB",    # secondary WAN (USB modem / phone)
    "WIRED" : "WIRED",      # wired connections
    "BRIDGE" : "BRIDGE",    # bridge
    "WIRELESS0" : "WLAN0",  # 2.4 GHz WiFi
    "WIRELESS1" : "WLAN1",  # 5 GHz WiFi
    "WIRELESS2" : "WLAN2",  # 5 GHz WiFi #2 (<-- check)
}
AR_KEY_WAN = "wanlink_state"
AR_KEY_WAN_STATE = (
    Key("wanstate", "state"),
    Key("wansbstate", "bstate"),
    Key("wanauxstate", "aux",),
    Key("autodet_state"),
    Key("autodet_auxstate",),
    Key("wanlink_status", "status", method = bool_from_any),
    Key("wanlink_type", "ip_type"),
    Key("wanlink_ipaddr", "ip"),
    Key("wanlink_netmask", "mask"),
    Key("wanlink_gateway", "gateway"),
    Key("wanlink_dns", "dns"),
    Key("wanlink_lease", "lease"),
    Key("wanlink_expires", "expires"),
    Key("is_private_subnet", "private_subnet", method = bool_from_any),
    Key("wanlink_xtype", "xtype"),
    Key("wanlink_xipaddr", "xip"),
    Key("wanlink_xnetmask", "xmask"),
    Key("wanlink_xgateway", "xgateway"),
    Key("wanlink_xdns", "xdns"),
    Key("wanlink_xlease", "xlease",),
    Key("wanlink_xexpires", "xexpires"),
)

AR_HOOK_TEMPLATE = "{}({});"
AR_HOOK_DEVICES = "{}()".format(AR_KEY_DEVICES)

AR_SERVICE_CONTROL : dict[str, list[str]] = {
    "firewall": [
        "restart",
    ],
    "ftpd": [
        "force_restart",
        "start",
        "stop",
        "restart",
    ],
    "httpd": [
        "restart",
    ],
    "wireless": [
        "restart",
    ],
}

AR_SERVICE_COMMAND : dict[str, str] = {
    "force_restart": "restart_{}_force",
    "start": "start_{}",
    "stop": "stop_{}",
    "restart": "restart_{}",
}

AR_SERVICE_DROP_CONNECTION : list[str] = {
    "httpd",
}

AR_DEFAULT_CORES = [ 1 ]
AR_DEFAULT_CORES_RANGE = range(1, 8)

DEFAULT_USAGE_DIGITS = 2


DEFAULT_CACHE_TIME = 5

DEFAULT_SLEEP_RECONNECT = 5
DEFAULT_SLEEP_CONNECTING = 1
DEFAULT_SLEEP_TIME = 0.1

DEFAULT_PORT = {
    "http": 80,
    "https": 8443,
}

AR_KEY = {
    "cpu_usage": "cpu_usage",
}

AR_PATH = {
    "command": "applyapp.cgi",
    "devicemap": "ajax_status.xml",
    "get": "appGet.cgi",
    "login": "login.cgi",
    "logout": "Logout.asp",
    "ports": "ajax_ethernet_ports.asp",
}

AR_ERROR = {
    "authorisation": 2,
    "credentials": 3,
    "try_again": 7,
    "logout": 8,
}

ERROR_IDENTITY = "Cannot obtain identity from the device {}. Exception summary{}"
ERROR_PARSING = "Failed parsing value '{}'. Please report this issue. Exception summary: {}"
ERROR_SERVICE = "Error calling service '{}'. Service did not return any expected value in the reply: {}"
ERROR_SERVICE_UNKNOWN = "Unknown service '{}' with mode '{}'"
ERROR_VALUE = "Wrong value '{}' with original exception: {}"
ERROR_VALUE_TYPE = "Wrong value '{}' of type {}"
ERROR_ZERO_DIVISION = "Zero division allert: {}"

MSG_ERROR = {
    "authorisation": "Session is not authorised",
    "cert_missing": "CA certificate is missing",
    "command": "Error sending a command {} to {}",
    "credentials": "Wrong credentials",
    "disabled_control": "Device is connected in no-control mode. Sending commands is blocked",
    "disabled_monitor": "Device is connected in no-monitor mode. Sending hooks is blocked",
    "token": "Cannot recieve a token from device",
    "try_again": "Too many attempts",
    "unknown": "Unknown error code {}",
}

MSG_WARNING = {
    "not_connected": "Not connected",
    "refused": "Connection refused",
}

MSG_INFO = {
    "drop_connection_service": "Connection will be dropped on behalf of '{}' service",
    "error_flag": "Error flag found",
    "empty_reqquest": "This request is empty",
    "identifying": "Identifying the device",
    "json_fix": "Trying to fix JSON response",
    "monitor_sleep": "Monitor {} is already active -> sleep",
    "monitor_wakeup": "Monitor {} woke up -> closing",
    "no_cert": "No certificate provided. Using trusted",
    "no_cert_check": "Certificate won't be checked",
    "reconnect": "Reconnecting",
    "service": "Service '{}' was called with arguments '{}' successfully. Reply: {}",
    "xml": "Data is in XML",
}

MSG_SUCCESS = {
    "cert_found": "CA certificate found",
    "command": "Command was sent successfully",
    "hook": "Hook was sent successfully",
    "identity": "Identity collected",
    "load": "Page {} was loaded successfully",
    "login": "Login successful",
    "logout": "Logout successful",
}

INTERFACE_TYPE = {
    "wan_ifnames": "wan",
    "wl_ifnames": "wlan",
    "wl0_vifnames": "gwlan2",
    "wl1_vifnames": "gwlan5",
    "lan_ifnames": "lan"
}

NVRAM_LIST = {
    "DEVICE": ["productid", "serial_no", "label_mac", "model", "pci/1/1/ATE_Brand", ],
    "MODE": ["sw_mode", "wlc_psta", "wlc_express", ],
    "FIRMWARE": ["firmver", "buildno", "extendno", ],
    "FUNCTIONS": ["rc_support", "ss_support", ],
    "INTERFACES": ["wans_dualwan", "wan_ifnames", "wl_ifnames", "wl0_vifnames", "wl1_vifnames", "lan_ifnames", "br0_ifnames", ],
    "REBOOT": ["reboot_schedule", "reboot_schedule_enable", "reboot_time", ],
    "WAN0": ["link_wan", "wan0_auxstate_t", "wan0_dns", "wan0_enable", "wan0_expires", "wan0_gateway", "wan0_gw_ifname", "wan0_gw_mac", "wan0_hwaddr", "wan0_ipaddr", "wan0_is_usb_modem_ready", "wan0_primary", "wan0_proto", "wan0_realip_ip", "wan0_realip_state", "wan0_sbstate_t", "wan0_state_t", "wan0_upnp_enable", ],
    "WAN1": ["link_wan1", "wan1_auxstate_t", "wan1_dns", "wan1_enable", "wan1_gateway", "wan1_gw_ifname", "wan1_gw_mac", "wan1_hwaddr", "wan1_ipaddr", "wan1_is_usb_modem_ready", "wan1_primary", "wan1_proto", "wan1_realip_ip", "wan1_realip_state", "wan1_sbstate_t", "wan1_state_t", "wan1_upnp_enable", ],
    "LAN": ["lan_gateway", "lan_hwaddr", "lan_ifname", "lan_ifnames", "lan_ipaddr", "lan_netmask", "lan_proto", "lanports", ],
    "WLAN0": ["wl0_radio", "wl0_ssid", "wl0_chanspec", "wl0_closed", "wl0_nmode_x", "wl0_bw", "wl0_auth_mode_x", "wl0_crypto", "wl0_wpa_psk", "wl0_mfp", "wl0_mbo_enable", "wl0_country_code", "wl0_maclist_x", "wl0_macmode", ],
    "WLAN1": ["wl1_radio", "wl1_ssid", "wl1_chanspec", "wl1_closed", "wl1_nmode_x", "wl1_bw", "wl1_auth_mode_x", "wl1_crypto", "wl1_wpa_psk", "wl1_mfp", "wl1_mbo_enable", "wl1_country_code", "wl1_maclist_x", "wl1_macmode", ],
    "GWLAN01": ["wl0.1_bss_enabled", "wl0.1_ssid", "wl0.1_auth_mode_x", "wl0.1_crypto", "wl0.1_key", "wl0.1_wpa_psk", "wl0.1_lanaccess", "wl0.1_expire", "wl0.1_expire_tmp", "wl0.1_macmode", "wl0.1_mbss", "wl0.1_sync_mode", ],
    "GWLAN02": ["wl0.2_bss_enabled", "wl0.2_ssid", "wl0.2_auth_mode_x", "wl0.2_crypto", "wl0.2_key", "wl0.2_wpa_psk", "wl0.2_lanaccess", "wl0.2_expire", "wl0.2_expire_tmp", "wl0.2_macmode", "wl0.2_mbss", "wl0.2_sync_mode", ],
    "GWLAN03": ["wl0.3_bss_enabled", "wl0.3_ssid", "wl0.3_auth_mode_x", "wl0.3_crypto", "wl0.3_key", "wl0.3_wpa_psk", "wl0.3_lanaccess", "wl0.3_expire", "wl0.3_expire_tmp", "wl0.3_macmode", "wl0.3_mbss", "wl0.3_sync_mode", ],
    "GWLAN11": ["wl1.1_bss_enabled", "wl1.1_ssid", "wl1.1_auth_mode_x", "wl1.1_crypto", "wl1.1_key", "wl1.1_wpa_psk", "wl1.1_lanaccess", "wl1.1_expire", "wl1.1_expire_tmp", "wl1.1_macmode", "wl1.1_mbss", "wl1.1_sync_mode", ],
    "GWLAN12": ["wl1.2_bss_enabled", "wl1.2_ssid", "wl1.2_auth_mode_x", "wl1.2_crypto", "wl1.2_key", "wl1.2_wpa_psk", "wl1.2_lanaccess", "wl1.2_expire", "wl1.2_expire_tmp", "wl1.2_macmode", "wl1.2_mbss", "wl1.2_sync_mode", ],
    "GWLAN13": ["wl1.3_bss_enabled", "wl1.3_ssid", "wl1.3_auth_mode_x", "wl1.3_crypto", "wl1.3_key", "wl1.3_wpa_psk", "wl1.3_lanaccess", "wl1.3_expire", "wl1.3_expire_tmp", "wl1.3_macmode", "wl1.3_mbss", "wl1.3_sync_mode", ],
}

MONITOR_MAIN = {
    "cpu_usage" : "appobj",
    "memory_usage" : "appobj",
    "netdev" : "appobj",
    "wanlink_state" : "appobj",
}

PORT_TYPE = [
    "LAN",
    "WAN",
]

KEY_NVRAM_GET = "nvram_get"
KEY_ACTION_MODE = "action_mode"
KEY_HOOK = "hook"
DEFAULT_ACTION_MODE = "apply"


### DEVICEMAP

# These values are just stored directly in this order in the corresponding node
DEVICEMAP_BY_INDEX = {
    "WAN": { "wan": ["status", "sbstatus", "auxstatus", ], },
    "WAN0": { "first_wan": ["status", "sbstatus", "auxstatus", ], },
    "WAN1": { "second_wan": [ "status", "sbstatus", "auxstatus", ], },
    "USB": { "usb": [ "status", ], },
}


#Not implemented yet
#psta:wlc_state=0;wlc_state_auth=0;
DEVICEMAP_SPECIAL = {
    "WAN": {
        "psta": [
            "wlc_state",
            "wlc_state_auth",
        ],
    },
}

# These values are stored as "key=value"
DEVICEMAP_GENERAL = {
    "WAN": {
        "wan": [ "monoClient", "wlc_state", "wlc_sbstate", "wifi_hw_switch", "ddnsRet", "ddnsUpdate", "wan_line_state", "wlan0_radio_flag", "wlan1_radio_flag", "wlan2_radio_flag", "data_rate_info_2g", "data_rate_info_5g", "data_rate_info_5g_2", "wan_diag_state", "active_wan_unit", "wlc0_state", "wlc1_state", "rssi_2g", "rssi_5g", "rssi_5g_2", "link_internet", "wlc2_state", "le_restart_httpd", ],
    },
    "VPN": {
        "vpn": [ "vpnc_proto", "vpnc_state_t", "vpnc_sbstate_t", "vpn_client1_state", "vpn_client2_state", "vpn_client3_state", "vpn_client4_state", "vpn_client5_state", "vpnd_state", "vpn_client1_errno", "vpn_client2_errno", "vpn_client3_errno", "vpn_client4_errno", "vpn_client5_errno", ],
    },
    "SYS": {
        "sys": [ "uptimeStr", ],
    },
    "QTN": {
        "qtn": [ "qtn_state", ],
    },
    "USB": {
        "usb": [ "modem_enable", ],
    },
    "WAN0": {
        "wan": [ "wan0_enable", "wan0_realip_state", "wan0_ipaddr", "wan0_realip_ip", ],
    },
    "WAN1": {
        "wan": [ "wan1_enable", "wan1_realip_state", "wan1_ipaddr", "wan1_realip_ip", ],
    },
    "SIM": {
        "sim": [ "sim_state", "sim_signal", "sim_operation", "sim_isp", "roaming", "roaming_imsi", "sim_imsi", "g3err_pin", "pin_remaining_count", "modem_act_provider", "rx_bytes", "tx_bytes", "modem_sim_order", ],
    },
    "DHCP": {
        "dhcp": [ "dnsqmode", ],
    },
    "DIAG": {
        "diag": [ "diag_dblog_enable", "diag_dblog_remaining", ]
    }
}

# This should be cleared from the data
DEVICEMAP_CLEAR = {
    "WAN": {
        "data_rate_info_2g": "\"",
        "data_rate_info_5g": "\"",
        "data_rate_info_5g_2": "\"",
    }
}





"""Library init"""

from .error import (
    AsusRouterConnectionError,
    AsusRouterConnectionTimeoutError,
    AsusRouterError,
    AsusRouterIdentityError,
    AsusRouterLoginBlockError,
    AsusRouterLoginError,
    AsusRouterNotImplementedError,
    AsusRouterResponseError,
    AsusRouterServerDisconnectedError,
    AsusRouterServiceError,
    AsusRouterSSLError,
    AsusRouterValueError,
)

from .dataclass import (
    AsusDevice as AsusDevice,
    ConnectedDevice as ConnectedDevice,
    Key as Key,
    Monitor as Monitor,
)

from .connection import (
    Connection as Connection,
)

from .asusrouter import (
    AsusRouter as AsusRouter,
)


"""Parsers module for AsusRouter"""

from __future__ import annotations

import logging
_LOGGER = logging.getLogger(__name__)

from typing import Any
import re
from datetime import datetime, timedelta
import xmltodict
import json

from asusrouter.const import(
    AR_DEFAULT_CORES,
    AR_DEFAULT_CORES_RANGE,
    AR_DEVICE_ATTRIBUTES_LIST,
    AR_KEY_CPU_ITEM,
    AR_KEY_CPU_LIST,
    AR_KEY_DEVICEMAP,
    AR_KEY_RAM_ITEM,
    AR_KEY_RAM_LIST,
    AR_KEY_NETWORK_ITEM,
    AR_KEY_NETWORK_GROUPS,
    AR_KEY_WAN_STATE,
    CONST_BITSINBYTE,
    CONST_ZERO,
    DATA_ADD_SPEED,
    DATA_TOTAL,
    DATA_TRAFFIC,
    DEFAULT_TRAFFIC_OVERFLOW,
    DEVICEMAP_BY_INDEX,
    DEVICEMAP_CLEAR,
    DEVICEMAP_GENERAL,
    ERROR_PARSING,
    ERROR_VALUE,
    KEY_NETWORK,
)
from asusrouter.dataclass import ConnectedDevice
from asusrouter.error import AsusRouterNotImplementedError, AsusRouterValueError
from asusrouter.util import calculators



def cpu_cores(raw : dict[str, Any] | None = None) -> list[int]:
    """CPU cores parser"""

    cores = list()

    if raw is None:
        return cores

    for i in AR_DEFAULT_CORES_RANGE:
        if any(AR_KEY_CPU_ITEM.format(i, data_type) in raw for data_type in AR_KEY_CPU_LIST):
            cores.append(i)
        else:
            break

    return cores


def cpu_usage(raw : dict[str, Any], cores : list[int] = AR_DEFAULT_CORES) -> dict[str, Any]:
    """CPU usage parser"""

    cpu = dict()

    # Populate total
    cpu[DATA_TOTAL] = dict()
    for item in AR_KEY_CPU_LIST:
        cpu[DATA_TOTAL][item.get()] = CONST_ZERO

    # Data / core
    for core in cores:
        cpu[core] = dict()

        for item in AR_KEY_CPU_LIST:
            key = AR_KEY_CPU_ITEM.format(core, item)
            new_key = item.get()
            if key in raw:
                try:
                    cpu[core][new_key] = int(raw[key])
                except ValueError as ex:
                    raise(AsusRouterValueError(ERROR_VALUE.format(raw[key], str(ex))))
                # Add this to total as well
                cpu[DATA_TOTAL][new_key] += cpu[core][new_key]

    return cpu


def ram_usage(raw : dict[str, Any]) -> dict[str, Any]:
    """RAM usage parser"""

    ram = dict()

    for item in AR_KEY_RAM_LIST:
        if AR_KEY_RAM_ITEM.format(item) in raw:
            try:
                ram[item] = int(raw[AR_KEY_RAM_ITEM.format(item)])
            except ValueError as ex:
                raise(AsusRouterValueError(ERROR_VALUE.format(raw[AR_KEY_RAM_ITEM.format(item)], str(ex))))

    return ram


def network_usage(raw : dict[str, Any], cache : dict[str, Any] | None = None) -> dict[str, Any]:
    """Network usage parser"""

    network = dict()
    for group in AR_KEY_NETWORK_GROUPS:
        for type in DATA_TRAFFIC:
            if AR_KEY_NETWORK_ITEM.format(group, type) in raw:
                if not AR_KEY_NETWORK_GROUPS[group] in network:
                    network[AR_KEY_NETWORK_GROUPS[group]] = dict()
                try:
                    network[AR_KEY_NETWORK_GROUPS[group]][type] = int(raw[AR_KEY_NETWORK_ITEM.format(group, type)], base = 16)
                except ValueError as ex:
                    raise(AsusRouterValueError(ERROR_VALUE.format(raw[AR_KEY_NETWORK_ITEM.format(group, type)], str(ex))))

            elif (cache is not None
                and KEY_NETWORK in cache
                and AR_KEY_NETWORK_GROUPS[group] in cache[DATA_TRAFFIC]
                and type in cache[DATA_TRAFFIC][AR_KEY_NETWORK_GROUPS[group]]
            ):
                network[AR_KEY_NETWORK_GROUPS[group]][type] = cache[DATA_TRAFFIC][AR_KEY_NETWORK_GROUPS[group]][type]

    return network


def network_speed(after : dict[str, dict[str, float]], before : dict[str, dict[str, float]], time_delta : float) -> dict[str, dict[str, float]]:
    """
    Network speed calculator

    Parameters
    -----
    `after`: current values. Outer dictionary `(groups)` contains inner dictionary `(types)`. On each `(type)` calculations are performed

    `before`: previous values

    `time_delta`: time between measurements

    Returns
    -----
    `after` with speeds append to `(types)`
    """

    for group in after:
        if group in before:
            speed = dict()
            for type in after[group]:
                if type in before[group]:
                    speed[DATA_ADD_SPEED.format(type)] = CONST_BITSINBYTE * calculators.speed(after = after[group][type], before = before[group][type], time_delta = time_delta, overflow = DEFAULT_TRAFFIC_OVERFLOW)
                else:
                    speed[DATA_ADD_SPEED.format(type)] = CONST_ZERO
            after[group] |= speed

    return after


def wan_state(raw : dict[str, Any]) -> dict[str, Any]:
    """WAN status parser"""

    values = dict()

    for key in AR_KEY_WAN_STATE:
        if (key.value in raw
            and raw[key.value] != str()
        ):
            try:
                values[key.get()] = key.method(raw[key.value]) if key.method else raw[key.value]
            except AsusRouterValueError as ex:
                _LOGGER.warning(ERROR_PARSING.format(key.value, str(ex)))

    return values


def connected_device(raw : dict[str, Any]) -> ConnectedDevice:
    """Device parser"""

    values = dict()

    for key in AR_DEVICE_ATTRIBUTES_LIST:
        if (key.value in raw
            and raw[key.value] != str()
        ):
            try:
                values[key.get()] = key.method(raw[key.value]) if key.method else raw[key.value]
            except AsusRouterValueError as ex:
                _LOGGER.warning(ERROR_PARSING.format(key.value, str(ex)))

    device = ConnectedDevice(**values)

    return device


def uptime(data : str) -> datetime:
    """Uptime -> boot time parser"""

    try:
        part = data.split("(")
        seconds = int(re.search("([0-9]+)", part[1]).group())
        when = datetime.strptime(part[0], "%a, %d %b %Y %H:%M:%S %z")
        boot = when - timedelta(seconds = seconds)
    except ValueError as ex:
        raise(AsusRouterValueError(ERROR_VALUE.format(data, str(ex))))

    return boot


def port_speed(value : str | None = None) -> int | None:
    """Port speed -> Mb/s parcer"""

    if value is None:
        return None
    elif value == "X":
        return 0
    elif value == "M":
        return 100
    elif value == "G":
        return 1000
    elif value == "Q":
        return 2500
    else:
        raise(AsusRouterNotImplementedError(value))


def devicemap(devicemap : dict[str, Any]) -> dict[str, Any]:
    """Devicemap parser"""

    data = {}

    # Get values only with index
    for node in DEVICEMAP_BY_INDEX:
        _node = {}
        for key in DEVICEMAP_BY_INDEX[node]:
            for el in range(len(DEVICEMAP_BY_INDEX[node][key])):
                _node[DEVICEMAP_BY_INDEX[node][key][el]] = devicemap[key][el]
            del devicemap[key][0 : (len(DEVICEMAP_BY_INDEX[node][key]) - 1)]
        data[node] = _node

    # Get values by key
    for node in DEVICEMAP_GENERAL:
        _node = {}
        for key in DEVICEMAP_GENERAL[node]:
            for el in DEVICEMAP_GENERAL[node][key]:
                if type(devicemap[key]) == str:
                    if el in devicemap[key]:
                        _node[el] = devicemap[key].replace("{}=".format(el), "")
                        break
                else:
                    for value in devicemap[key]:
                        if el in value:
                            _node[el] = value.replace("{}=".format(el), "")
                            break
        if node in data:
            data[node].update(_node)
        else:
            data[node] = _node

    # Clear values from useless symbols
    for node in DEVICEMAP_CLEAR:
        for value in DEVICEMAP_CLEAR[node]:
            data[node][value] = data[node][value].replace(DEVICEMAP_CLEAR[node][value], "")

    return data


def pseudo_json(text : str) -> dict[str, Any]:
    """JSON parser"""

    data = re.sub('\s+','',text)

    if "get_wan_lan_status=" in data:
        data = data.replace("get_wan_lan_status=", "")
        data = data[:-1]
    elif "varcpuInfo,memInfo=newObject();cpuInfo=" in data:
        data = data.replace("varcpuInfo,memInfo=newObject();cpuInfo=", '{"cpu":{')
        data = data.replace(";memInfo=", '},"memory":{')
        data = data.replace(";", "}}")
    else:
        _LOGGER.error("Unknown data. Template for this data is unknown")
        return {}
    
    return json.loads(data.encode().decode('utf-8-sig') )


def xml(text : str) -> dict[str, Any]:
    """XML parser"""

    data = xmltodict.parse(text)
    if AR_KEY_DEVICEMAP in data:
        return devicemap(data[AR_KEY_DEVICEMAP])

    return {}





"""Error module"""

from __future__ import annotations

from typing import Any, Optional
import aiohttp
import asyncio


class AsusRouterError(Exception):
    """Base class for errors in AsusRouter library"""

    def __init__(self, *args : Any, message : Optional[str] = None, **_kwargs : Any) -> None:
        """Initialise base error"""

        super().__init__(*args, message)


class AsusRouterConnectionError(AsusRouterError, aiohttp.ClientConnectionError):
    """Error connecting to the router"""


class AsusRouterConnectionTimeoutError(AsusRouterError, aiohttp.ServerTimeoutError, asyncio.TimeoutError):
    """Timeout error on communication"""


class AsusRouterServerDisconnectedError(AsusRouterError, aiohttp.ServerDisconnectedError):
    """Server disconnected error"""


class AsusRouterSSLError(AsusRouterError, aiohttp.ClientConnectorSSLError, aiohttp.ClientConnectorCertificateError):
    """SSL error"""


class AsusRouterLoginError(AsusRouterError):
    """Login error / credentials error"""


class AsusRouterLoginBlockError(AsusRouterError):
    """Too many attempts error on device side"""

    def __init__(self, *args : Any, message : Optional[str] = None, timeout : Optional[int] = None, **_kwargs : Any) -> None:
        """Initialise base error"""

        self._timeout = timeout

        super().__init__(*args, message)

    
    @property
    def timeout(self) -> int | None:
        """Return timeout"""

        return self._timeout


class AsusRouterResponseError(AsusRouterError, aiohttp.ClientResponseError):
    """Error on communication"""


class AsusRouterServiceError(AsusRouterError):
    """Error on calling a service"""


class AsusRouterValueError(AsusRouterError, ValueError):
    """Invalid value received"""


class AsusRouterNotImplementedError(AsusRouterError, NotImplementedError):
    """Not implemented error"""


class AsusRouterIdentityError(AsusRouterError):
    """Error of collecting device identity"""


